@@ -20,7 +20,7 @@
  *     mark_dummy_rel()
  *     restriction_is_constant_false()
  *
- * Portions Copyright (c) 1996-2014, PostgreSQL Global Development Group
+ * Portions Copyright (c) 1996-2015, PostgreSQL Global Development Group
  * Portions Copyright (c) 1994, Regents of the University of California
  *
  *-------------------------------------------------------------------------
@@ -215,9 +215,6 @@ set_append_rel_pathlist(PlannerInfo *root, RelOptInfo *rel,
 			add_path(rel, (Path *)
 					 create_append_path(rel, subpaths, required_outer));
 	}
-
-	/* Select cheapest paths */
-	set_cheapest(rel);
 }
 
 /*
@@ -720,7 +717,7 @@ join_search_one_level(PlannerInfo *root, int level)
 		 */
 		if (joinrels[level] == NIL &&
 			root->join_info_list == NIL &&
-			root->lateral_info_list == NIL)
+			!root->hasLateralRTEs)
 			elog(ERROR, "failed to build any %d-way joins", level);
 	}
 }
@@ -1047,15 +1044,7 @@ join_is_legal(PlannerInfo *root, RelOptInfo *rel1, RelOptInfo *rel2,
 				 match_sjinfo->jointype == JOIN_FULL))
 				return false;	/* not implementable as nestloop */
 			/* check there is a direct reference from rel2 to rel1 */
-			foreach(l, root->lateral_info_list)
-			{
-				LateralJoinInfo *ljinfo = (LateralJoinInfo *) lfirst(l);
-
-				if (bms_is_subset(ljinfo->lateral_rhs, rel2->relids) &&
-					bms_is_subset(ljinfo->lateral_lhs, rel1->relids))
-					break;
-			}
-			if (l == NULL)
+			if (!bms_overlap(rel1->relids, rel2->direct_lateral_relids))
 				return false;	/* only indirect refs, so reject */
 			/* check we won't have a dangerous PHV */
 			if (have_dangerous_phv(root, rel1->relids, rel2->lateral_relids))
@@ -1070,15 +1059,7 @@ join_is_legal(PlannerInfo *root, RelOptInfo *rel1, RelOptInfo *rel2,
 				 match_sjinfo->jointype == JOIN_FULL))
 				return false;	/* not implementable as nestloop */
 			/* check there is a direct reference from rel1 to rel2 */
-			foreach(l, root->lateral_info_list)
-			{
-				LateralJoinInfo *ljinfo = (LateralJoinInfo *) lfirst(l);
-
-				if (bms_is_subset(ljinfo->lateral_rhs, rel1->relids) &&
-					bms_is_subset(ljinfo->lateral_lhs, rel2->relids))
-					break;
-			}
-			if (l == NULL)
+			if (!bms_overlap(rel2->relids, rel1->direct_lateral_relids))
 				return false;	/* only indirect refs, so reject */
 			/* check we won't have a dangerous PHV */
 			if (have_dangerous_phv(root, rel2->relids, rel1->lateral_relids))
@@ -1141,7 +1122,6 @@ join_is_legal(PlannerInfo *root, RelOptInfo *rel1, RelOptInfo *rel2,
 	return true;
 }
 
-
 /*
  * has_join_restriction
  *		Detect whether the specified relation has join-order restrictions,
